// $Id: AST.cpp,v 1.10 2007-11-22 06:12:43 babic Exp $

// Copyright (C) 2007, Domagoj Babic, see the LICENSE.

#include "AST.h"

#include <sstream>
#include <iterator>

extern void yyerror(const char *str);

using namespace std;
using namespace spear_tools;

namespace {

// JS Hash Function 
static unsigned int JSHash(const std::string& str) {
   unsigned int hash = 1315423911;
   for (std::string::const_iterator i = str.begin(); i != str.end(); i++) { 
      hash ^= ((hash << 5) + (unsigned)*i + (hash >> 2));
   }
   return hash;
}

// Robert Jenkins' 32 bit Mix Function
static unsigned int RJHash (const unsigned int key) {
    unsigned int hash = key;
    hash += (hash << 12);
    hash ^= (hash >> 22);
    hash += (hash << 4);
    hash ^= (hash >> 9);
    hash += (hash << 10);
    hash ^= (hash >> 2);
    hash += (hash << 7);
    hash ^= (hash >> 12);
    return hash;
}

} // End of anonymous namespace

void AstNode::dump() const { 
    dump(std::cerr); 
}

Expression::Expression(const string &Op, astv_iterator B, astv_iterator
        E) : AstNode(OP_TY, 1), Op(Op) {

    const unsigned sz = distance(B, E);
    if (sz == 2) {
        Successors.push_back(*B);
        ++B;
        Successors.push_back(*B);
    } else if (sz == 3) {
        Successors.push_back(*B);
        AstVec tmp(++B, E);
        Successors.push_back(new Expression(Op, tmp));
    } else {
        Successors.push_back(new Expression(Op, B, B + sz/2));
        Successors.push_back(new Expression(Op, B + sz/2, E));
    }

    assert(getSuccessor(0)->numBits() == 1 &&
           getSuccessor(1)->numBits() == 1 &&
           "This constructor can be called only for predicates.");
}

Expression::Expression(const string &Op, AstVec &Ops) :
    AstNode(Ops, 0), Op(Op) {

    unsigned sz = Ops.size();

    // Compute the number of bits of the expression

    if (Op == "&" || Op == "|" || Op == "^") {
        if (sz == 2) {
            goto Binary_operators;
        } else if (sz < 50) {
            // SMT lib allows and to have > 2 operators

            // First clear successors
            Successors.clear();
            AstVec().swap(Successors);
            while (Ops.size() != 2) {
                AstVec S(1, Ops.back());
                Ops.pop_back();
                S.push_back(Ops.back());
                Ops.pop_back();
                Ops.push_back(new Expression(Op, S));
            }

            copy(Ops.begin(), Ops.end(), back_inserter(Successors));
            sz = Successors.size();
            goto Binary_operators;
        } else {
            // First clear successors
            Successors.clear();
            AstVec().swap(Successors);
            astv_iterator I = Ops.begin();
            Successors.push_back(new Expression(Op, I, I + sz/2));
            Successors.push_back(new Expression(Op, I + sz/2, Ops.end()));
            sz = Successors.size();
            goto Binary_operators;
        }
    } else if (Op == "+"   || Op == "-"  || Op == "*"   ||
	    Op == "/s"  || Op == "/u" || Op == "%s"  ||
	    Op == "%u"  || Op == "<<" || Op == ">>a" ||
	    Op == ">>l" || Op == "~") {

        if (Op == "~") {
            if (sz != 1) { 
                cerr << *Ops[0] << "   " << *Ops[1] << endl;
                yyerror("not requires a single operand."); 
            }
        } else {
Binary_operators:
            if (sz != 2) { 
                string Msg("operator " + Op + " requires 2 operands");
                yyerror(Msg.c_str());
            }
            if (Op == "<<" || Op == ">>a" || Op == ">>l") {
                assert(Ops[1]->numBits() == 8 && "Invalid num of shift bits.");
            } else if (Ops[0]->numBits() != Ops[1]->numBits()) {
                string Msg("operator " + Op + 
                    " requires 2 operands of equal type");
                yyerror(Msg.c_str());
            }
        }
        
        // Result always has the number of bits equal to the first
        // operand.
        bits = Ops[0]->numBits();
    } else if (
        Op == "="   || Op == "/="  || Op == "ule" ||
	    Op == "uge" || Op == "ult" || Op == "ugt" ||
	    Op == "sle" || Op == "sge" || Op == "slt" ||
	    Op == "sgt" || Op == "=>") {

        if (sz != 2) {
            string Msg("operator " + Op + " requires 2 operands");
            yyerror(Msg.c_str());
        }
        if (Ops[0]->numBits() != Ops[1]->numBits()) {
            std::cerr << *Ops[0] << "    " << *Ops[1] << endl;
            string Msg("operator " + Op + 
                " requires 2 operands of equal type");
            yyerror(Msg.c_str());
        }
        
        // Predicates always produce a single-bit result
        bits = 1;
    } else if (Op == "ite") {
        if (sz != 3) {
            string Msg("operator " + Op + " requires 3 operands");
            yyerror(Msg.c_str());
        }
        if (Ops[0]->numBits() != 1) {
            string Msg("operator " + Op + 
                " requires the first operand to be boolean");
            yyerror(Msg.c_str());
        }
        if (Ops[1]->numBits() != Ops[2]->numBits()) {
            string Msg("operator " + Op + 
                " requires the last 2 operands of equal type");
            yyerror(Msg.c_str());
        }
        
        // The result has equal number of bits as the second operand.
        bits = Ops[1]->numBits();
    } else if (Op == "sext" || Op == "trun" || Op == "zext") {
        if (sz != 2) {
            string Msg("operator " + Op + " requires 2 operands");
            yyerror(Msg.c_str());
        }
        // No need to set the number of bits here. This expression is
        // always associated with a 

        // The second parameter should be a constant containing the
        // resulting number of bits (originally sign_extend contains the
        // number of preprended bits).
        if (const Constant *C = Ops[1]->asCon()) {
            if (Op == "trun") {
                bits = C->value();
            } else {
                bits = Ops[0]->numBits() + C->value();
            }
        } else {
            assert(false && "Second param of sext not a const.");
        }

        // Finally discard the last parameter because SF format requires
        // only one.
        Successors.pop_back();
    } else if (Op == "conc") {
        if (sz != 2) {
            string Msg("operator " + Op + " requires 2 operands");
            yyerror(Msg.c_str());
        }
        
        // Sum the bitwidths of both params.
        bits = Ops[0]->numBits() + Ops[1]->numBits();
	} else if (Op == "extr") {
        if (sz != 3) {
            string Msg("operator " + Op + " requires 3 operands");
            yyerror(Msg.c_str());
        }

        // Both params should be constants, store their diff as the
        // number of bits (originally extract contains [x:y] inclusive
        // bounds).
        const Constant *C1 = Ops[1]->asCon(), *C2 = Ops[2]->asCon();

        if (C1 && C2) {
            const unsigned b1 = C1->value(), b2 = C2->value();
            if (b2 <= b1) {
                yyerror("second range constant must be strictly larger");
            }
            if (b2 > Ops[0]->numBits()) {
                cerr << *this << "(" << *Ops[0] << "," << *Ops[1] <<
                    "," << *Ops[2] << ")" << endl;
                yyerror("second range constant is indexing out of bounds");
            }
            bits = b2 - b1;
        } else {
            yyerror("the last two params of extr should be constants");
        }
    } else {
        assert(false && "Unsupported operator.");
    }

    assert(bits > 0 && "Invalid number of bits.");

    // Update hash
    Hash = hash();
}

unsigned Expression::NxtTmp(0);

unsigned Expression::hash() const {
    const unsigned sz = numSuccessors();
    unsigned hsh = JSHash(Op) ^ RJHash(numBits());
    for (unsigned i = 0; i < sz; ++i) {
        hsh -= getSuccessor(i)->getHash();
    }
    return hsh;
}

void Expression::generate(ostream &OS, TableTy &SymbolTable,
        bool first) {

    typedef vector<pair<string, AstNode*> > ToDoTy;
    typedef ToDoTy::iterator todo_iterator;

    ToDoTy *td = new ToDoTy();

    if (first) {
        OS << "p ";
    }
    OS << Op << " ";
    
    for (const_iterator I = begin(), E = end(); I != E; ++I) {
        AstNode *Suc = *I;
        if (Suc->asExpr()) {
            string Nm;

            {
                ostringstream nm;
                do {
                    nm.str("");
                    nm << "Tmp" << (++NxtTmp);
                } while (SymbolTable.find(nm.str()) != SymbolTable.end());
                Nm = nm.str();
                SymbolTable.insert(make_pair(Nm, Suc));
            }

            // Dump operand
            OS << Nm;
            // Dump definition
            {
                td->push_back(make_pair(Nm, Suc));
            }
            // Dump declaration
            cout << Nm << ":i" << Suc->numBits() << " ";
        } else {
            Suc->generate(OS, SymbolTable, false);
        }
        OS << " ";
    }

    OS << endl;
    for (todo_iterator I = td->begin(), E = td->end(); I != E; ++I) {
        AstNode *S = I->second;
        OS << "c " << I->first << " ";
        I->second->generate(OS, SymbolTable, false);
    }
    delete td;
}

Variable *Expression::getFreshVar(unsigned bits) {
    stringstream ss;
    ss << "Tmp" << (++NxtTmp);
    return new Variable(ss.str(), bits);
}

Variable *Expression::getFreshVar(AstNode *Def) {
    stringstream ss;
    ss << "Tmp" << (++NxtTmp);
    return new Variable(ss.str(), Def);
}

void Expression::dump(std::ostream &os) const {
    os << Op << ":i" << numBits();
}

unsigned Constant::hash() const {
    return RJHash(static_cast<unsigned>(val)) ^ RJHash(numBits());
}

void Constant::generate(ostream &OS, TableTy &SymbolTable, bool first) {
    if (first) {
        OS << "p = 1:i1 ";
    }

    OS << val << ":i" << numBits();

    if (first) {
        OS << endl;
    }
}

void Constant::dump(std::ostream &os) const {
    os << val << ":i" << numBits();
}

unsigned Variable::hash() const {
    if (numSuccessors() == 1) {
        return getSuccessor(0)->getHash();
    } else {
        return JSHash(Nm) ^ RJHash(numBits());
    }
}

bool Variable::operator==(const AstNode &A) const {
    const Variable *V = A.asVar();
    if (!V) { 
        // Can't be equal if the node types differ
        return false; 
    }

    const unsigned Sucs = numSuccessors();
    if (Sucs && Sucs == V->numSuccessors()) {
        assert(Sucs == 1 && "Variable has multiple definitions?");
        // Variables can be renamed, always compare definitions.

        const AstNode *Suc = getSuccessor(0);
        const AstNode *VSuc = V->getSuccessor(0);

        return getHash() == V->getHash() && *Suc == *VSuc;
    } else {
        return V->Nm == Nm && V->numBits() == numBits();
    }
}

void Variable::generate(ostream &OS, TableTy &SymbolTable, bool first) {
    // This call will not add any prefix, but will cleanup
    // disallowed characters.
    prefix("");

    if (first) {
        if (hasSuccessors()) {
            assert(numSuccessors() == 1 && 
                "Unexpected number of successors.");
            if (getSuccessor(0)->asExpr()) {
                OS << "c " << Nm << " ";
                getSuccessor(0)->generate(OS, SymbolTable, false);
                // :formula (flet ($def (not a)) $def))
                // is added twice as a conjunction, and both times this
                // constraint is generated:
                // c fvar0_def ~ a
                // Addition of the following line assures that the
                // second time predicate:
                // p = 1:i1 fvar0_de
                // is generated. This bug was found and fixed by 
                // Alberto Griggio (thx!).
                const_cast<AstVec &>(getSucVec()).clear();
            } else {
                OS << "p = " << Nm << " ";
                getSuccessor(0)->generate(OS, SymbolTable, false);
                OS << endl;
            }
        } else {
            OS << "p = 1:i1 " << Nm << endl;
        }
    } else {
        OS << Nm;
    }
}

void Variable::prefix(const string &Pre) {
    string::size_type idx = 0;
    while ((idx = Nm.find_first_of("'.$?", idx)) != string::npos) {
        Nm[idx] = '_';
    }
    Nm = Pre + Nm;

    idx = Nm.find_first_not_of(
    "0123456789_ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz");
    if (idx != string::npos) {
        cerr << "Error: unsupported character: " << Nm[idx] << endl;
        assert(false && "Unsupported character in SF var name.");
    }
}

void Variable::dump(std::ostream &os) const {
    os << Nm << ":i" << numBits();
}
