// $Id: AST.h,v 1.8 2007-11-22 06:11:16 babic Exp $

// Copyright (C) 2007, Domagoj Babic, see the LICENSE.

#ifndef ASTNODE_H
#define ASTNODE_H

#include <vector>
#include <iostream>
#include <string>
#include <cassert>
#include <map>

namespace spear_tools {

class AstNode;
class Expression;
class Constant;
class Variable;

typedef std::vector<AstNode*> AstVec;
typedef AstVec::iterator astv_iterator;
typedef std::vector<Constant*> AstConstVec;

typedef std::map<std::string, AstNode*> TableTy;
typedef TableTy::iterator symbol_iterator;

class AstNode {
protected:
    enum NodeTy { VAR_TY, CON_TY, OP_TY};

    // Operands of operators, or definitions of variables
    AstVec Successors;
    // Type of the node
    NodeTy NTy;
    // Number of bits required for the representation of the value.
    unsigned bits;
    // Hash
    unsigned Hash;

    template <class T>
    inline T *get(NodeTy Ty) {
        return (Ty == NTy) ? static_cast<T*>(this) : NULL;
    }
    
    template <class T>
    inline const T *get(NodeTy Ty) const {
        return (Ty == NTy) ? static_cast<const T*>(this) : NULL;
    }

public:

    typedef AstVec::iterator iterator;
    typedef AstVec::const_iterator const_iterator;

    iterator begin() { return Successors.begin(); }
    iterator end() { return Successors.end(); }
    const_iterator begin() const { return Successors.begin(); }
    const_iterator end() const { return Successors.end(); }

    // Constructor for expressions
    AstNode(const AstVec &Suc, unsigned Bits) : 
        Successors(Suc.begin(), Suc.end()), 
        NTy(OP_TY), bits(Bits) {}

    // Constructor for vars and constants
    AstNode(NodeTy NTy, unsigned Bits) : NTy(NTy), bits(Bits) {} 

    virtual ~AstNode() {}

    // Hashing is used solely to speed up node comparison - if the
    // hashes don't match, the nodes definitely represent different
    // expressions.
    unsigned getHash() const { return Hash; }

    // Operation on successors
    unsigned numSuccessors() const { return Successors.size(); }
    bool hasSuccessors() const { return !Successors.empty(); }
    AstNode *getSuccessor(unsigned i ) {
        assert(i < numSuccessors() && "Index out of bounds.");
        return Successors[i];
    }
    const AstNode *getSuccessor(unsigned i ) const {
        assert(i < numSuccessors() && "Index out of bounds.");
        return Successors[i];
    }
    const AstVec &getSucVec() const { return Successors; }

    // Recursively generate SF format. Constants return unsigned
    // constant as string, variables return the variable name, while
    // operators return encoding "op operands", which has to be named
    // with a temporary.
    virtual void generate(std::ostream &OS, 
			TableTy &SymbolTable, bool first) = 0;
    
    // Equality
    virtual bool operator==(const AstNode&) const = 0;

    // Cast support
    Expression *asExpr() { return get<Expression>(OP_TY); }
    Constant *asCon() { return get<Constant>(CON_TY); }
    Variable *asVar() { return get<Variable>(VAR_TY); }

    // And their const versions
    const Expression *asExpr() const { return get<Expression>(OP_TY); }
    const Constant *asCon() const { return get<Constant>(CON_TY); }
    const Variable *asVar() const { return get<Variable>(VAR_TY); }

    // Returns the type of the node (number of bits).
    unsigned numBits() const { 
        assert(bits > 0 && "Invalid number of bits.");
        return bits; 
    }

    // Debug
    virtual void dump(std::ostream &os) const = 0;
    void dump() const;
};

inline std::ostream &operator<<(std::ostream &os, const AstNode &N) {
    N.dump(os); return os;
}

inline bool operator==(const AstVec &A, const AstVec &B) {
    const unsigned sz = A.size();
    if (sz != B.size()) { return false; }

    for (unsigned i = 0; i < sz; ++i) {
        const AstNode *a = A[i], *b = B[i];
        if (a == b) { 
            return true;
        }
        if (a->getHash() != b->getHash() || !(*a == *b)) {
            return false;
        }
    }
    return true;
}

class Expression : public AstNode {
    std::string Op;
    static unsigned NxtTmp;
    
	Expression(const std::string &Op, astv_iterator B, astv_iterator E);

protected:

    unsigned hash() const;

public:
    Expression(const std::string &Op, AstVec &Ops);

    inline bool operator==(const AstNode &A) const {
        const Expression *E = A.asExpr();
        return E && E->getHash() == getHash() && 
            E->Op == Op && Successors == A.getSucVec();
    }

    void generate(std::ostream &OS, TableTy &SymbolTable, 
			bool first);

    std::string op() const { return Op; }
    static Variable *getFreshVar(unsigned bits);
    static Variable *getFreshVar(AstNode *Def);

    // Debug
    void dump(std::ostream &os) const;
};

class Constant : public AstNode {
    const uint64_t val;

protected:

    unsigned hash() const;

public:
    Constant(uint64_t val, unsigned Bits) :
        AstNode(CON_TY, Bits), val(val) { 
            assert(Bits > 0 && "Invalid number of bits.");
            Hash = hash(); 
        }
    
    inline bool operator==(const AstNode &A) const {
        const Constant *C = A.asCon();
        return C && C->val == val && C->numBits() == bits;
    }

    void generate(std::ostream &OS, TableTy &SymbolTable,
			bool first);

    uint64_t value() const { return val; }
    void setNumBits(unsigned Bits) { bits = Bits; }
    
    // Debug
    void dump(std::ostream &os) const;
};

class Variable : public AstNode {
    std::string Nm;

protected:

    unsigned hash() const;

public:
    Variable(const std::string &Name, unsigned Bits) :
        AstNode(VAR_TY, Bits), Nm(Name) { 
            Hash = hash(); 
            assert(Bits > 0 && "Invalid number of bits.");
        }
    Variable(const std::string &Name, AstNode *Def) :
        AstNode(VAR_TY, Def->numBits()), Nm(Name) { 
            assert(bits > 0 && "Invalid number of bits.");
            Successors.push_back(Def);
            Hash = hash();
        }
    
    bool operator==(const AstNode &A) const;

    void generate(std::ostream &OS, TableTy &SymbolTable, bool first);
    // Replaces characters not supported by Spear ('.$?) with an
    // underline, and prepends prefix. Fails if the resulting name
    // either does not start with a letter, or contains an unsupported
    // character.
    void prefix(const std::string &Pre);

    std::string name() const { return Nm; }
    
    // Debug
    void dump(std::ostream &os) const;
};

} // End of spear_tools namespace

#endif // ASTNODE_H
